/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * Copyright (c) 1998,1999 Thomas Roessler <roessler@guug.de>
 * Copyright (C) 2002 g10 Code GmbH
 * 
 *     This program is free software; you can redistribute it
 *     and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation; either
 *     version 2 of the License, or (at your option) any later
 *     version.
 * 
 *     This program is distributed in the hope that it will be
 *     useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more
 *     details.
 * 
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 59 Temple Place - Suite 330,
 *     Boston, MA  02111, USA.
 */

#include "mutt.h"

#ifdef HAVE_GPGME

#include "mutt_curses.h"
#include "mutt_menu.h"
#include "mime.h"
#include "crypt.h"
#include "pager.h"
#include "sort.h"

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <locale.h>

/* safe_free is evil in that a cast is required and one often forgets
   to pass the address and not just the pointer.  This scheme bypasses
   all type checking and thus is not reqyure.  And free() is suppored
   to be a noop when a NULL pointer is passed.  A macro is the correct
   solution. */
#undef FREE
#define FREE(a) do { free (a); a = NULL; } while (0)
#define safe_free(a) you_should_not_use_safe_free (a);


/* some macros to replace ctype ones and avoid locale problems */
#define spacep(p)   (*(p) == ' ' || *(p) == '\t')
#define digitp(p)   (*(p) >= '0' && *(p) <= '9')
#define hexdigitp(a) (digitp (a)                     \
                      || (*(a) >= 'A' && *(a) <= 'F')  \
                      || (*(a) >= 'a' && *(a) <= 'f'))
/* the atoi macros assume that the buffer has only valid digits */
#define atoi_1(p)   (*(p) - '0' )
#define atoi_2(p)   ((atoi_1(p) * 10) + atoi_1((p)+1))
#define atoi_4(p)   ((atoi_2(p) * 100) + atoi_2((p)+2))
#define xtoi_1(p)   (*(p) <= '9'? (*(p)- '0'): \
                     *(p) <= 'F'? (*(p)-'A'+10):(*(p)-'a'+10))
#define xtoi_2(p)   ((xtoi_1(p) * 16) + xtoi_1((p)+1))



/* Values used for comparing addresses. */
#define CRYPT_KV_VALID    1
#define CRYPT_KV_ADDR     2
#define CRYPT_KV_STRING   4
#define CRYPT_KV_STRONGID 8
#define CRYPT_KV_MATCH (CRYPT_KV_ADDR|CRYPT_KV_STRING)


struct crypt_cache
{
  char *what;
  char *dflt;
  struct crypt_cache *next;
};


struct dn_array_s {
  char *key;
  char *value;
};

struct crypt_entry
{
  size_t num;
  crypt_key_t *key;
};
typedef struct crypt_entry crypt_entry_t;


static struct crypt_cache *id_defaults = NULL;

const char *crypt_pkalgbytype (unsigned char);


/* Return the keyID for the key K.  Note that this string is valid as
   long as K is valid */
const char *crypt_keyid (crypt_key_t *k)
{
  const char *s;

  s = gpgme_key_get_string_attr (k->kobj, GPGME_ATTR_KEYID, NULL, 0);
  if (!s)
    return "????????"; /* oops*/
  if ( !option (OPTPGPLONGIDS) && strlen (s) == 16)
    s += 8; /* return only the short keyID */
  return s;
}

/* Return the hexstring fingerprint from the key K. */
const char *crypt_fpr (crypt_key_t *k)
{
  const char *s;

  s = gpgme_key_get_string_attr (k->kobj, GPGME_ATTR_FPR, NULL, 0);
  if (!s)
    return ""; /* oops */
  return s;
}


/* Parse FLAGS and return a statically allocated(!) string with them. */
static char *crypt_key_abilities (int flags)
{
  static char buff[3];

  if (!(flags & KEYFLAG_CANENCRYPT))
    buff[0] = '-';
  else if (flags & KEYFLAG_PREFER_SIGNING)
    buff[0] = '.';
  else
    buff[0] = 'e';

  if (!(flags & KEYFLAG_CANSIGN))
    buff[1] = '-';
  else if (flags & KEYFLAG_PREFER_ENCRYPTION)
    buff[1] = '.';
  else
    buff[1] = 's';

  buff[2] = '\0';

  return buff;
}

/* Parse FLAGS and return a character describing the most important flag. */
static char crypt_flags (int flags)
{
  if (flags & KEYFLAG_REVOKED)
    return 'R';
  else if (flags & KEYFLAG_EXPIRED)
    return 'X';
  else if (flags & KEYFLAG_DISABLED)
    return 'd';
  else if (flags & KEYFLAG_CRITICAL)
    return 'c';
  else 
    return ' ';
}

/* Return a copy of KEY. */
static crypt_key_t *crypt_copy_key (crypt_key_t *key)
{
  crypt_key_t *k;

  k = safe_calloc (1, sizeof *k);
  k->kobj = key->kobj;
  gpgme_key_ref (key->kobj);
  k->idx = key->idx;
  k->uid = key->uid;
  k->flags = key->flags;

  return k;
}

/* Release all the keys at the address of KEYLIST and set the address
   to NULL. */
void crypt_free_key (crypt_key_t **keylist)
{
  while (*keylist)
    {
      crypt_key_t *k = (*keylist)->next;
      FREE (k);
      *keylist = k;
    }
}



/*
 * Format an entry on the CRYPT key selection menu.
 * 
 * %n	number
 * %k	key id		%K 	key id of the principal key
 * %u	user id
 * %a	algorithm	%A      algorithm of the princ. key
 * %l	length		%L	length of the princ. key
 * %f	flags		%F 	flags of the princ. key
 * %c	capabilities	%C	capabilities of the princ. key
 * %t	trust/validity of the key-uid association
 * %[...] date of key using strftime(3)
 */

static const char *crypt_entry_fmt (char *dest,
                                    size_t destlen,
                                    char op,
                                    const char *src,
                                    const char *prefix,
                                    const char *ifstring,
                                    const char *elsestring,
                                    unsigned long data,
                                    format_flag flags)
{
  char fmt[16];
  crypt_entry_t *entry;
  crypt_key_t *key;
  int kflags = 0;
  int optional = (flags & M_FORMAT_OPTIONAL);
  const char *s;
  unsigned long val;

  entry = (crypt_entry_t *) data;
  key   = entry->key;

/*    if (isupper ((unsigned char) op)) */
/*      key = pkey; */

  kflags = (key->flags /*| (pkey->flags & KEYFLAG_RESTRICTIONS)
                         | uid->flags*/);
  
  switch (ascii_tolower (op))
    {
    case '[':
      {
	const char *cp;
	char buf2[SHORT_STRING], *p;
	int do_locales;
	struct tm *tm;
	size_t len;

	p = dest;

	cp = src;
	if (*cp == '!')
	{
	  do_locales = 0;
	  cp++;
	}
	else
	  do_locales = 1;

	len = destlen - 1;
	while (len > 0 && *cp != ']')
	{
	  if (*cp == '%')
	  {
	    cp++;
	    if (len >= 2)
	    {
	      *p++ = '%';
	      *p++ = *cp;
	      len -= 2;
	    }
	    else
	      break; /* not enough space */
	    cp++;
	  }
	  else
	  {
	    *p++ = *cp++;
	    len--;
	  }
	}
	*p = 0;

	if (do_locales && Locale)
	  setlocale (LC_TIME, Locale);
        
        {
          time_t tt = (time_t)gpgme_key_get_ulong_attr (key->kobj,
                                                        GPGME_ATTR_CREATED,
                                                        0, 0);
          tm = localtime (&tt);
        }
	strftime (buf2, sizeof (buf2), dest, tm);

	if (do_locales)
	  setlocale (LC_TIME, "C");
        
	snprintf (fmt, sizeof (fmt), "%%%ss", prefix);
	snprintf (dest, destlen, fmt, buf2);
	if (len > 0)
	  src = cp + 1;
      }
      break;
    case 'n':
      if (!optional)
      {
	snprintf (fmt, sizeof (fmt), "%%%sd", prefix);
	snprintf (dest, destlen, fmt, entry->num);
      }
      break;
    case 'k':
      if (!optional)
      {
        /* fixme: we need a way to distinguish between main and subkeys.
           Store the idx in entry? */
	snprintf (fmt, sizeof (fmt), "%%%ss", prefix);
	snprintf (dest, destlen, fmt, crypt_keyid (key));
      }
      break;
    case 'u':
      if (!optional)
      {
	snprintf (fmt, sizeof (fmt), "%%%ss", prefix);
	snprintf (dest, destlen, fmt, key->uid);
      }
      break;
    case 'a':
      if (!optional)
      {
	snprintf (fmt, sizeof (fmt), "%%%s.3s", prefix);
        s = gpgme_key_get_string_attr (key->kobj, GPGME_ATTR_ALGO,
                                       NULL, 0);
        if (!s)
          s = "?";
	snprintf (dest, destlen, fmt, s);
      }
      break;
    case 'l':
      if (!optional)
      {
	snprintf (fmt, sizeof (fmt), "%%%slu", prefix);
        val = gpgme_key_get_ulong_attr (key->kobj, GPGME_ATTR_LEN,
                                        NULL, 0);
	snprintf (dest, destlen, fmt, val);
      }
      break;
    case 'f':
      if (!optional)
      {
	snprintf (fmt, sizeof (fmt), "%%%sc", prefix);
	snprintf (dest, destlen, fmt, crypt_flags (kflags));
      }
      else if (!(kflags & (KEYFLAG_RESTRICTIONS)))
        optional = 0;
      break;
    case 'c':
      if (!optional)
      {
	snprintf (fmt, sizeof (fmt), "%%%ss", prefix);
	snprintf (dest, destlen, fmt, crypt_key_abilities (kflags));
      }
      else if (!(kflags & (KEYFLAG_ABILITIES)))
        optional = 0;
      break;
    case 't':
      if ((kflags & KEYFLAG_ISX509))
        s = "x";
      else
        s = gpgme_key_get_string_attr (key->kobj, GPGME_ATTR_VALIDITY,
                                       NULL, key->idx);
      snprintf (fmt, sizeof (fmt), "%%%sc", prefix);
      snprintf (dest, destlen, fmt, s? *s: 'B');
      break;
    default:
      *dest = '\0';
  }

  if (optional)
    mutt_FormatString (dest, destlen, ifstring, mutt_attach_fmt, data, 0);
  else if (flags & M_FORMAT_OPTIONAL)
    mutt_FormatString (dest, destlen, elsestring, mutt_attach_fmt, data, 0);
  return (src);
}
      
/* Used by the display fucntion to format a line. */
static void crypt_entry (char *s, size_t l, MUTTMENU * menu, int num)
{
  crypt_key_t **key_table = (crypt_key_t **) menu->data;
  crypt_entry_t entry;
  
  entry.key = key_table[num];
  entry.num = num + 1;

  mutt_FormatString (s, l, NONULL (PgpEntryFormat), crypt_entry_fmt, 
		     (unsigned long) &entry, M_FORMAT_ARROWCURSOR);
}

/* Compare two addresses and the keyid to be used for sorting. */
static int _crypt_compare_address (const void *a, const void *b)
{
  crypt_key_t **s = (crypt_key_t **) a;
  crypt_key_t **t = (crypt_key_t **) b;
  int r;

  if ((r = mutt_strcasecmp ((*s)->uid, (*t)->uid)))
    return r > 0;
  else
    return mutt_strcasecmp (crypt_keyid (*s), crypt_keyid (*t)) > 0;
}

static int crypt_compare_address (const void *a, const void *b)
{
  return ((PgpSortKeys & SORT_REVERSE) ? !_crypt_compare_address (a, b)
                                         :  _crypt_compare_address (a, b));
}


/* Compare two key IDs and the addresses to be used for sorting. */
static int _crypt_compare_keyid (const void *a, const void *b)
{
  crypt_key_t **s = (crypt_key_t **) a;
  crypt_key_t **t = (crypt_key_t **) b;
  int r;

  if ((r = mutt_strcasecmp (crypt_keyid (*s), crypt_keyid (*t))))
    return r > 0;
  else
    return mutt_strcasecmp ((*s)->uid, (*t)->uid) > 0;
}

static int crypt_compare_keyid (const void *a, const void *b)
{
  return ((PgpSortKeys & SORT_REVERSE) ? !_crypt_compare_keyid (a, b)
		                         :  _crypt_compare_keyid (a, b));
}

/* Compare 2 creation dates and the addresses.  For sorting. */
static int _crypt_compare_date (const void *a, const void *b)
{
  crypt_key_t **s = (crypt_key_t **) a;
  crypt_key_t **t = (crypt_key_t **) b;
  unsigned long ts, tt;
  
  ts = gpgme_key_get_ulong_attr ((*s)->kobj, GPGME_ATTR_CREATED, NULL, 0);
  tt = gpgme_key_get_ulong_attr ((*t)->kobj, GPGME_ATTR_CREATED, NULL, 0);
  if (ts > tt)
    return 1;
  if (ts < tt)
    return 0;
  return mutt_strcasecmp ((*s)->uid, (*t)->uid) > 0;
}

static int crypt_compare_date (const void *a, const void *b)
{
  return ((PgpSortKeys & SORT_REVERSE) ? !_crypt_compare_date (a, b)
                                         :  _crypt_compare_date (a, b));
}

/* Compare two trust values, the key length, the creation dates. the
   addresses and the key IDs.  For sorting. */
static int _crypt_compare_trust (const void *a, const void *b)
{
  crypt_key_t **s = (crypt_key_t **) a;
  crypt_key_t **t = (crypt_key_t **) b;
  int r;
  unsigned long ts, tt;

  if ((r = (((*s)->flags & (KEYFLAG_RESTRICTIONS))
	    - ((*t)->flags & (KEYFLAG_RESTRICTIONS)))))
    return r > 0;
  ts = gpgme_key_get_ulong_attr ((*s)->kobj, GPGME_ATTR_VALIDITY, NULL, 0);
  tt = gpgme_key_get_ulong_attr ((*t)->kobj, GPGME_ATTR_VALIDITY, NULL, 0);
  if ((r = (tt - ts)))
    return r < 0;
  ts = gpgme_key_get_ulong_attr ((*s)->kobj, GPGME_ATTR_LEN, NULL, 0);
  tt = gpgme_key_get_ulong_attr ((*t)->kobj, GPGME_ATTR_LEN, NULL, 0);
  if (ts != tt)
    return ts > tt;
  ts = gpgme_key_get_ulong_attr ((*s)->kobj, GPGME_ATTR_CREATED, NULL, 0);
  tt = gpgme_key_get_ulong_attr ((*t)->kobj, GPGME_ATTR_CREATED, NULL, 0);
  if (ts > tt)
    return 1;
  if (ts < tt)
    return 0;
  if ((r = mutt_strcasecmp ((*s)->uid, (*t)->uid)))
    return r > 0;
  return (mutt_strcasecmp (crypt_keyid ((*s)), crypt_keyid ((*t)))) > 0;
}

static int crypt_compare_trust (const void *a, const void *b)
{
  return ((PgpSortKeys & SORT_REVERSE) ? !_crypt_compare_trust (a, b)
  				       : _crypt_compare_trust (a, b));
}

/* Return trute when key K is valid. */
static int crypt_key_is_valid (crypt_key_t *k)
{
  if (k->flags & KEYFLAG_CANTUSE)
    return 0;
  return 1;
}

/* Return true whe validity of KEY is sufficient. */
static int crypt_id_is_strong (crypt_key_t *key)
{
  GpgmeValidity val;

  if ((key->flags & KEYFLAG_ISX509))
    return 1;
  
  val = gpgme_key_get_ulong_attr (key->kobj, GPGME_ATTR_VALIDITY,
                                 NULL, key->idx);
  switch (val)
    {
    case GPGME_VALIDITY_UNKNOWN:   return 0;
    case GPGME_VALIDITY_UNDEFINED: return 0;
    case GPGME_VALIDITY_NEVER:     return 0;
    case GPGME_VALIDITY_MARGINAL:  return 0;
    case GPGME_VALIDITY_FULL:      return 1;
    case GPGME_VALIDITY_ULTIMATE:  return 1;
    }
  return 0;
}

/* Return true when the KEY is valid, i.e. not marked as unusable. */
static int crypt_id_is_valid (crypt_key_t *key)
{
  if (key->flags & KEYFLAG_CANTUSE)
    return 0;
  /* else */
  return 1;
}

/* Return a bit vector describing how well the addresses ADDR and
   U_ADDR match and whether KEY is valid. */
static int crypt_id_matches_addr (ADDRESS *addr, ADDRESS *u_addr,
                                  crypt_key_t *key)
{
  int rv = 0;
  
  if (crypt_id_is_valid (key))
    rv |= CRYPT_KV_VALID;

  if (crypt_id_is_strong (key))
    rv |= CRYPT_KV_STRONGID;
  
  if (addr->mailbox && u_addr->mailbox
      && mutt_strcasecmp (addr->mailbox, u_addr->mailbox) == 0)
    rv |= CRYPT_KV_ADDR;
    
  if (addr->personal && u_addr->personal
      && mutt_strcasecmp (addr->personal, u_addr->personal) == 0)
    rv |= CRYPT_KV_STRING;
  
  return rv;
}

/* return true when S pints to a didgit or letter. */
static int
digit_or_letter (const unsigned char *s)
{
  return ( (*s >= '0' && *s < '9')
           || (*s >= 'A' && *s <= 'Z')
           || (*s >= 'a' && *s <= 'z'));
}


/* Print the utf-8 encoded string BUF of length LEN bytes to stream
   FP. Convert the character set. */
static void
print_utf8 (FILE *fp, const char *buf, size_t len)
{
  char *tstr;

  tstr = safe_malloc (len+1);
  memcpy (tstr, buf, len);
  tstr[len] = 0;
  mutt_convert_string (&tstr, "utf-8", Charset, M_ICONV_HOOK_FROM);
  fputs (tstr, fp);
  FREE (tstr);
}

/* Print the X.500 Distinguished Name part KEY from the array of parts
   DN to FP. */
static int
print_dn_part (FILE *fp, struct dn_array_s *dn, const char *key)
{
  int any = 0;

  for (; dn->key; dn++)
    {
      if (!strcmp (dn->key, key))
        {
          if (any)
            fputs (" + ", fp);
          print_utf8 (fp, dn->value, strlen (dn->value));
          any = 1;
        }
    }
  return any;
}

/* Print all parts of a DN in a standard sequence. */
static void
print_dn_parts (FILE *fp, struct dn_array_s *dn)
{
  const char *stdpart[] = {
    "CN", "OU", "O", "STREET", "L", "ST", "C", NULL 
  };
  int any=0, any2=0, i;
  
  for (i=0; stdpart[i]; i++)
    {
      if (any)
        fputs (", ", fp);
      any = print_dn_part (fp, dn, stdpart[i]);
    }
  /* now print the rest without any specific ordering */
  for (; dn->key; dn++)
    {
      for (i=0; stdpart[i]; i++)
        {
          if (!strcmp (dn->key, stdpart[i]))
            break;
        }
      if (!stdpart[i])
        {
          if (any)
            fputs (", ", fp);
          if (!any2)
            fputs ("(", fp);
          any = print_dn_part (fp, dn, dn->key);
          any2 = 1;
        }
    }
  if (any2)
    fputs (")", fp);
}


/* Parse an RDN; this is a helper to parse_dn(). */
static const unsigned char *
parse_dn_part (struct dn_array_s *array, const unsigned char *string)
{
  const unsigned char *s, *s1;
  size_t n;
  unsigned char *p;

  /* parse attributeType */
  for (s = string+1; *s && *s != '='; s++)
    ;
  if (!*s)
    return NULL; /* error */
  n = s - string;
  if (!n)
    return NULL; /* empty key */
  array->key = safe_malloc (n+1);
  p = (unsigned char *)array->key;
  memcpy (p, string, n); /* fixme: trim trailing spaces */
  p[n] = 0;
  string = s + 1;

  if (*string == '#')
    { /* hexstring */
      string++;
      for (s=string; hexdigitp (s); s++)
        s++;
      n = s - string;
      if (!n || (n & 1))
        return NULL; /* empty or odd number of digits */
      n /= 2;
      p = safe_malloc (n+1);
      array->value = (char*)p;
      for (s1=string; n; s1 += 2, n--)
        *p++ = xtoi_2 (s1);
      *p = 0;
   }
  else
    { /* regular v3 quoted string */
      for (n=0, s=string; *s; s++)
        {
          if (*s == '\\')
            { /* pair */
              s++;
              if (*s == ',' || *s == '=' || *s == '+'
                  || *s == '<' || *s == '>' || *s == '#' || *s == ';' 
                  || *s == '\\' || *s == '\"' || *s == ' ')
                n++;
              else if (hexdigitp (s) && hexdigitp (s+1))
                {
                  s++;
                  n++;
                }
              else
                return NULL; /* invalid escape sequence */
            }
          else if (*s == '\"')
            return NULL; /* invalid encoding */
          else if (*s == ',' || *s == '=' || *s == '+'
                   || *s == '<' || *s == '>' || *s == '#' || *s == ';' )
            break; 
          else
            n++;
        }

      p = safe_malloc (n+1);
      array->value = (char*)p;
      for (s=string; n; s++, n--)
        {
          if (*s == '\\')
            { 
              s++;
              if (hexdigitp (s))
                {
                  *p++ = xtoi_2 (s);
                  s++;
                }
              else
                *p++ = *s;
            }
          else
            *p++ = *s;
        }
      *p = 0;
    }
  return s;
}


/* Parse a DN and return an array-ized one.  This is not a validating
   parser and it does not support any old-stylish syntax; gpgme is
   expected to return only rfc2253 compatible strings. */
static struct dn_array_s *
parse_dn (const unsigned char *string)
{
  struct dn_array_s *array;
  size_t arrayidx, arraysize;
  int i;

  arraysize = 7; /* C,ST,L,O,OU,CN,email */
  array = safe_malloc ((arraysize+1) * sizeof *array);
  arrayidx = 0;
  while (*string)
    {
      while (*string == ' ')
        string++;
      if (!*string)
        break; /* ready */
      if (arrayidx >= arraysize)
        { /* mutt lacks a real safe_realoc - so we need to copy */
          struct dn_array_s *a2;

          arraysize += 5;
          a2 = safe_malloc ((arraysize+1) * sizeof *array);
          for (i=0; i < arrayidx; i++)
            {
              a2[i].key = array[i].key;
              a2[i].value = array[i].value;
            }
          FREE (array);
          array = a2;
        }
      array[arrayidx].key = NULL;
      array[arrayidx].value = NULL;
      string = parse_dn_part (array+arrayidx, string);
      arrayidx++;
      if (!string)
        goto failure;
      while (*string == ' ')
        string++;
      if (*string && *string != ',' && *string != ';' && *string != '+')
        goto failure; /* invalid delimiter */
      if (*string)
        string++;
    }
  array[arrayidx].key = NULL;
  array[arrayidx].value = NULL;
  return array;

 failure:
  for (i=0; i < arrayidx; i++)
    {
      FREE (array[i].key);
      FREE (array[i].value);
    }
  FREE (array);
  return NULL;
}


/* Print a nice representation of the USERID and make sure it is
   displayed in a proper way, which does mean to reorder some parts
   for S/MIME's DNs.  USERID is a string as returned by the gpgme key
   functions.  It is utf-8 encoded. */
static void
parse_and_print_user_id (FILE *fp, const char *userid)
{
  const char *s;
  int i;

  if (*userid == '<')
    {
      s = strchr (userid+1, '>');
      if (s)
        print_utf8 (fp, userid+1, s-userid-1);
    }
  else if (*userid == '(')
    fputs (_("[Can't display this user ID (unknown encoding)]"), fp);
  else if (!digit_or_letter ((const unsigned char *)userid))
    fputs (_("[Can't display this user ID (invalid encoding)]"), fp);
  else
    {
      struct dn_array_s *dn = parse_dn ((const unsigned char *)userid);
      if (!dn)
        fputs (_("[Can't display this user ID (invalid DN)]"), fp);
      else 
        {
          print_dn_parts (fp, dn);          
          for (i=0; dn[i].key; i++)
            {
              FREE (dn[i].key);
              FREE (dn[i].value);
            }
          FREE (dn);
        }
    }
}

/* Pirnt verbose information about a key or certificate to FP. */
static void print_key_info (GpgmeKey key, FILE *fp)
{
  int idx;
  const char *s, *s2;
  time_t tt;
  struct tm *tm;
  char shortbuf[SHORT_STRING];
  unsigned long aval;
  const char *delim;
  int is_pgp;
  int i;

  if (Locale)
    setlocale (LC_TIME, Locale);

  is_pgp = !gpgme_key_get_ulong_attr (key, GPGME_ATTR_TYPE, NULL, 0);
  for (idx=0; (s = gpgme_key_get_string_attr (key, GPGME_ATTR_USERID,
                                              NULL, idx)); idx++ )
    {
      if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_UID_REVOKED, NULL, idx))
        continue;
      fprintf (fp, "%s ......: ", idx? _(" aka"):_("Name"));
      if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_UID_INVALID, NULL, idx))
        {
          fputs (_("[Invalid]"), fp);
          putc (' ', fp);
        }
      if (is_pgp)
        print_utf8 (fp, s, strlen(s));
      else
        parse_and_print_user_id (fp, s);
      putc ('\n', fp);
    }


  tt = (time_t)gpgme_key_get_ulong_attr (key, GPGME_ATTR_CREATED, NULL, 0);
  if (tt)
    {
      tm = localtime (&tt);
      strftime (shortbuf, sizeof shortbuf, "%c", tm);
      fprintf (fp, "Valid From : %s\n", shortbuf);
    }

  tt = (time_t)gpgme_key_get_ulong_attr (key, GPGME_ATTR_EXPIRE, NULL, 0);
  if (tt)
    {
      tm = localtime (&tt);
      strftime (shortbuf, sizeof shortbuf, "%c", tm);
      fprintf (fp, "Valid To ..: %s\n", shortbuf);
    }
  s = gpgme_key_get_string_attr (key, GPGME_ATTR_ALGO, NULL, 0);
  if (!s)
    s = "?";
  s2 = gpgme_key_get_string_attr (key, GPGME_ATTR_TYPE, NULL, 0);
  if (!s2)
    s2 = "?";
  aval = gpgme_key_get_ulong_attr (key, GPGME_ATTR_LEN, NULL, 0);
  fprintf (fp, "Key Type ..: %s, %lu bit %s\n", s2, aval, s);

  fprintf (fp, "Key Usage .: ");
  delim = "";
  if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_CAN_ENCRYPT, NULL, 0))
    {
      fprintf (fp, "%s%s", delim, _("encryption"));
      delim = ", ";
    }
  if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_CAN_SIGN, NULL, 0))
    {
      fprintf (fp, "%s%s", delim, _("signing"));
      delim = ", ";
    }
  if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_CAN_CERTIFY, NULL, 0))
    {
      fprintf (fp, "%s%s", delim, _("certification"));
      delim = ", ";
    }
  putc ('\n', fp);

  s = gpgme_key_get_string_attr (key, GPGME_ATTR_FPR, NULL, 0);
  if (s)
    {
      fputs (_("Fingerprint: "), fp);
      if (is_pgp && strlen (s) == 40)
        {
          for (i=0; *s && s[1] && s[2] && s[3] && s[4]; s += 4, i++)
            {
              putc (*s, fp);
              putc (s[1], fp);
              putc (s[2], fp);
              putc (s[3], fp);
              putc (is_pgp? ' ':':', fp);
              if (is_pgp && i == 4)
                putc (' ', fp);
            }
        }
      else
        {
          for (i=0; *s && s[1] && s[2]; s += 2, i++)
            {
              putc (*s, fp);
              putc (s[1], fp);
              putc (is_pgp? ' ':':', fp);
              if (is_pgp && i == 7)
                putc (' ', fp);
            }
        }
      fprintf (fp, "%s\n", s);
    }

  s = gpgme_key_get_string_attr (key, GPGME_ATTR_SERIAL, NULL, 0);
  if (s)
    fprintf (fp, "Serial-No .: 0x%s\n", s);
  
  s = gpgme_key_get_string_attr (key, GPGME_ATTR_ISSUER, NULL, 0);
  if (s)
    {
      fprintf (fp, "Issued By .: ");
      parse_and_print_user_id (fp, s);
      putc ('\n', fp);
    }

  /* For PGP we list all subkeys. */
  if (is_pgp)
    {
      for (idx=1; (s = gpgme_key_get_string_attr (key, GPGME_ATTR_KEYID,
                                                  NULL, idx)); idx++ )
        {
          putc ('\n', fp);
          if ( strlen (s) == 16)
            s += 8; /* display only the short keyID */
          fprintf (fp, "Subkey ....: 0x%s", s);
          if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_KEY_REVOKED,
                                         NULL, idx))
            {
              putc (' ', fp);
              fputs (_("[Revoked]"), fp);
            }
          if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_KEY_INVALID,
                                         NULL, idx))
            {
              putc (' ', fp);
              fputs (_("[Invalid]"), fp);
            }
          if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_KEY_EXPIRED,
                                         NULL, idx))
            {
              putc (' ', fp);
              fputs (_("[Expired]"), fp);
            }
          if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_KEY_DISABLED,
                                         NULL, idx))
            {
              putc (' ', fp);
              fputs (_("[Disabled]"), fp);
            }
          putc ('\n', fp);

          tt = (time_t)gpgme_key_get_ulong_attr (key, GPGME_ATTR_CREATED,
                                                 NULL, idx);
          if (tt)
            {
              tm = localtime (&tt);
              strftime (shortbuf, sizeof shortbuf, "%c", tm);
              fprintf (fp, "Valid From : %s\n", shortbuf);
            }

          tt = (time_t)gpgme_key_get_ulong_attr (key, GPGME_ATTR_EXPIRE,
                                                 NULL, idx);
          if (tt)
            {
              tm = localtime (&tt);
              strftime (shortbuf, sizeof shortbuf, "%c", tm);
              fprintf (fp, "Valid To ..: %s\n", shortbuf);
            }
          s = gpgme_key_get_string_attr (key, GPGME_ATTR_ALGO, NULL, idx);
          if (!s)
            s = "?";
          aval = gpgme_key_get_ulong_attr (key, GPGME_ATTR_LEN, NULL, idx);
          fprintf (fp, "Key Type ..: %s, %lu bit %s\n", "PGP", aval, s);

          fprintf (fp, "Key Usage .: ");
          delim = "";
          s = gpgme_key_get_string_attr (key, GPGME_ATTR_KEY_CAPS, NULL, idx);
          if (s && strchr (s, 'e'))
            {
              fprintf (fp, "%s%s", delim, _("encryption"));
              delim = ", ";
            }
          if (s && strchr (s, 's'))
            {
              fprintf (fp, "%s%s", delim, _("signing"));
              delim = ", ";
            }
          if (s && strchr (s, 'c'))
            {
              fprintf (fp, "%s%s", delim, _("certification"));
              delim = ", ";
            }
          putc ('\n', fp);
        }
    }

  if (Locale)
    setlocale (LC_TIME, "C");
}


/* Show detailed information about the selected key */
static void 
verify_key (crypt_key_t *key)
{
  FILE *fp;
  char cmd[LONG_STRING], tempfile[_POSIX_PATH_MAX];
  const char *s;
  GpgmeCtx listctx = NULL;
  GpgmeError err;
  GpgmeKey k = NULL;
  int maxdepth = 100;

  mutt_mktemp (tempfile);
  if (!(fp = safe_fopen (tempfile, "w")))
    {
      mutt_perror _("Can't create temporary file");
      return;
    }
  mutt_message _("Collecting data...");

  print_key_info (key->kobj, fp);

  err = gpgme_new (&listctx);
  if (err)
    {
      fprintf (fp, "Internal error: can't create gpgme context: %s\n",
               gpgme_strerror (err));
      goto leave;
    }
  if ((key->flags & KEYFLAG_ISX509))
      gpgme_set_protocol (listctx, GPGME_PROTOCOL_CMS);

  k = key->kobj;
  gpgme_key_ref (k);
  while ((s=gpgme_key_get_string_attr (k, GPGME_ATTR_CHAINID, NULL, 0)))
    {
      putc ('\n', fp);
      err = gpgme_op_keylist_start (listctx, s, 0);
      gpgme_key_release (k);
      k = NULL;
      if (!err)
	err = gpgme_op_keylist_next (listctx, &k);
      if (err)
        {
          fprintf (fp, _("Error finding issuer key: %s\n"),
                   gpgme_strerror (err));
          goto leave;
        }
      gpgme_op_keylist_end (listctx);
      
      print_key_info (k, fp);
      if (!--maxdepth)
        {
          putc ('\n', fp);
          fputs (_("Error: certification chain to long - stopping here\n"),
                 fp);
          break;
        }
    }

 leave:
  gpgme_key_release (k);
  gpgme_release (listctx);
  fclose (fp);
  mutt_clear_error ();
  snprintf (cmd, sizeof (cmd), _("Key ID: 0x%s"),  crypt_keyid (key));
  mutt_do_pager (cmd, tempfile, 0, NULL);
}


/* Display a menu to select a key from the array KEYS. */
static crypt_key_t *crypt_select_key (crypt_key_t *keys,
                                      ADDRESS * p, const char *s, 
                                      unsigned int app)
{
  int keymax;
  crypt_key_t **key_table;
  MUTTMENU *menu;
  int i, done = 0;
  char helpstr[SHORT_STRING], buf[LONG_STRING];
  crypt_key_t *k;
  int (*f) (const void *, const void *);
  
  int unusable = 0;

  /* build the key table */
  keymax = i = 0;
  key_table = NULL;
  for (k = keys; k; k = k->next)
    {
      if (!option (OPTPGPSHOWUNUSABLE) && (k->flags & KEYFLAG_CANTUSE))
        {
          unusable = 1;
          continue;
        }
      
      if (i == keymax)
        {
          keymax += 20;
          safe_realloc ((void **) &key_table, sizeof (crypt_key_t*)*keymax);
        }
      
      key_table[i++] = k;
    }

  if (!i && unusable)
    {
      mutt_error _("All matching keys are marked expired/revoked.");
      mutt_sleep (1);
      return NULL;
    }

  switch (PgpSortKeys & SORT_MASK)
  {
    case SORT_DATE:
      f = crypt_compare_date;
      break;
    case SORT_KEYID:
      f = crypt_compare_keyid;
      break;
    case SORT_ADDRESS:
      f = crypt_compare_address;
      break;
    case SORT_TRUST:
    default:
      f = crypt_compare_trust;
      break;
  }
  qsort (key_table, i, sizeof (crypt_key_t*), f);

  helpstr[0] = 0;
  mutt_make_help (buf, sizeof (buf), _("Exit  "), MENU_CRYPT, OP_EXIT);
  strcat (helpstr, buf);	/* __STRCAT_CHECKED__ */
  mutt_make_help (buf, sizeof (buf), _("Select  "), MENU_CRYPT,
		  OP_GENERIC_SELECT_ENTRY);
  strcat (helpstr, buf);	/* __STRCAT_CHECKED__ */
  mutt_make_help (buf, sizeof (buf), _("Check key  "), MENU_CRYPT, OP_VERIFY_KEY);
  strcat (helpstr, buf);	/* __STRCAT_CHECKED__ */
  mutt_make_help (buf, sizeof (buf), _("Help"), MENU_CRYPT, OP_HELP);
  strcat (helpstr, buf);	/* __STRCAT_CHECKED__ */

  menu = mutt_new_menu ();
  menu->max = i;
  menu->make_entry = crypt_entry;
  menu->menu = MENU_CRYPT;
  menu->help = helpstr;
  menu->data = key_table;

  {
    const char *ts;

    if ((app & APPLICATION_PGP) && (app &  APPLICATION_SMIME))
      ts = _("PGP and S/MIME keys matching");
    else if ((app & APPLICATION_PGP))
      ts = _("PGP keys matching");
    else if ((app & APPLICATION_SMIME))
      ts = _("S/MIME keys matching");
    else
      ts = _("keys matching");

    if (p)
      snprintf (buf, sizeof (buf), _("%s <%s>."), ts, p->mailbox);
    else
      snprintf (buf, sizeof (buf), _("%s \"%s\"."), ts, s);
    menu->title = buf; 
  }

  mutt_clear_error ();
  k = NULL;
  while (!done)
    {
      switch (mutt_menuLoop (menu))
        {
        case OP_VERIFY_KEY:
          verify_key (key_table[menu->current]);
          menu->redraw = REDRAW_FULL;
          break;
          
        case OP_VIEW_ID:
          mutt_message ("%s", key_table[menu->current]->uid);
          break;
          
        case OP_GENERIC_SELECT_ENTRY:
          /* FIXME make error reporting more verbose - this should be
             easy because gpgme provides more information */
          if (option (OPTPGPCHECKTRUST))
            {
            if (!crypt_key_is_valid (key_table[menu->current]))
              {
                mutt_error _("This key can't be used: "
                             "expired/disabled/revoked.");
                break;
              }
            }
          
          if (option (OPTPGPCHECKTRUST) &&
              (!crypt_id_is_valid (key_table[menu->current])
               || !crypt_id_is_strong (key_table[menu->current])))
            {
              char *s = "";
              char buff[LONG_STRING];
              
              if (key_table[menu->current]->flags & KEYFLAG_CANTUSE)
                s = N_("ID is expired/disabled/revoked.");
              else 
                {
                  GpgmeValidity val;
                  
                  s = "??";
                  val = gpgme_key_get_ulong_attr
                    (key_table[menu->current]->kobj, GPGME_ATTR_VALIDITY,
                     NULL, key_table[menu->current]->idx);
                  switch (val)
                    {
                    case GPGME_VALIDITY_UNKNOWN:   
                    case GPGME_VALIDITY_UNDEFINED: 
                      s = N_("ID has undefined validity.");
                      break;
                    case GPGME_VALIDITY_NEVER:     
                      s = N_("ID is not valid.");
                      break;
                    case GPGME_VALIDITY_MARGINAL:  
                      s = N_("ID is only marginally valid.");
                      break;
                    case GPGME_VALIDITY_FULL:      
                    case GPGME_VALIDITY_ULTIMATE:  
                      break;
                    }
                  
                  snprintf (buff, sizeof (buff),
                            _("%s Do you really want to use the key?"),
                            _(s));
              
                  if (mutt_yesorno (buff, 0) != 1)
                    {
                      mutt_clear_error ();
                      break;
                    }
                }
            }  

          k = crypt_copy_key (key_table[menu->current]);
          done = 1;
          break;
          
        case OP_EXIT:
          k = NULL;
          done = 1;
          break;
        }
    }
  
  mutt_menuDestroy (&menu);
  FREE (key_table);

  set_option (OPTNEEDREDRAW);
  
  return k;
}


/* Display TAG as a prompt to ask for a key.  If WHATFOR is not null
   use it as default and store it under that label as the next
   default.  ABILITIES describe the required key abilities (sign,
   encrypt) and APP the type of the requetsed key; ether S/MIME or
   PGP.  Return a copy of the key or NULL if not found. */
crypt_key_t *crypt_ask_for_key (char *tag, char *whatfor, short abilities,
                                unsigned int app)
{
  crypt_key_t *key;
  char resp[SHORT_STRING];
  struct crypt_cache *l = NULL;

  mutt_clear_error ();
  
  resp[0] = 0;
  if (whatfor)
    {
      
      for (l = id_defaults; l; l = l->next)
        if (!mutt_strcasecmp (whatfor, l->what))
          {
            strfcpy (resp, NONULL (l->dflt), sizeof (resp));
            break;
          }
    }


  for (;;)
    {
      resp[0] = 0;
      if (mutt_get_field (tag, resp, sizeof (resp), M_CLEAR) != 0)
        return NULL;
      
      if (whatfor)
        {
          if (l)
            mutt_str_replace (&l->dflt, resp);
          else
            {
              l = safe_malloc (sizeof (struct crypt_cache));
              l->next = id_defaults;
              id_defaults = l;
              l->what = safe_strdup (whatfor);
              l->dflt = safe_strdup (resp);
            }
        }
      
      if ((key = crypt_getkeybystr (resp, abilities, app)))
        return key;
      
      BEEP ();
    }
  /* not reached */
}

/* Show information about the keys in string DUMMY. */
void crypt_view_key (const char *dummy)
{
  crypt_key_t *key;

  /* FIXME: This is currently just a test routine. we need to tell the
     pager  about the keys it is displaying. */
  key = crypt_ask_for_key ("Enter key pattern to show: ", NULL, 0,
                           APPLICATION_SMIME|APPLICATION_PGP);
  if (key)
    crypt_free_key (&key);
}

/* generate a public key attachment */
BODY *crypt_make_key_attachment (char *tempf)
{
#if 0
  BODY *att;
  char buff[LONG_STRING];
  char tempfb[_POSIX_PATH_MAX], tmp[STRING];
  FILE *tempfp;
  FILE *devnull;
  struct stat sb;
  pid_t thepid;
  crypt_key_t *key;
  unset_option (OPTPGPCHECKTRUST);

  key = crypt_ask_for_key (_("Please enter the key ID: "), NULL, 0);

  if (!key)    return NULL;

  snprintf (tmp, sizeof (tmp), "0x%s", crypt_keyid (key));
  crypt_free_key (&key);
  
  if (!tempf)
  {
    mutt_mktemp (tempfb);
    tempf = tempfb;
  }

  if ((tempfp = safe_fopen (tempf, tempf == tempfb ? "w" : "a")) == NULL)
  {
    mutt_perror _("Can't create temporary file");
    return NULL;
  }

  if ((devnull = fopen ("/dev/null", "w")) == NULL)	/* __FOPEN_CHECKED__ */
  {
    mutt_perror _("Can't open /dev/null");
    fclose (tempfp);
    if (tempf == tempfb)
      unlink (tempf);
    return NULL;
  }

  mutt_message _("Invoking pgp...");

  
  if ((thepid = 
       crypt_invoke_export (NULL, NULL, NULL, -1,
			   fileno (tempfp), fileno (devnull), tmp)) == -1)
  {
    mutt_perror _("Can't create filter");
    unlink (tempf);
    fclose (tempfp);
    fclose (devnull);
    return NULL;
  }

  mutt_wait_filter (thepid);

  fclose (tempfp);
  fclose (devnull);

  att = mutt_new_body ();
  att->filename = safe_strdup (tempf);
  att->unlink = 1;
  att->use_disp = 0;
  att->type = TYPEAPPLICATION;
  att->subtype = safe_strdup ("crypt-keys");
  snprintf (buff, sizeof (buff), _("CRYPT Key %s."), tmp);
  att->description = safe_strdup (buff);
  mutt_update_encoding (att);

  stat (tempf, &sb);
  att->length = sb.st_size;

  return att;
#endif
  return NULL;
}


/* Add the string STR to the list HINTS.  This list is later used to
   match addresses. */
static LIST *crypt_add_string_to_hints (LIST *hints, const char *str)
{
  char *scratch;
  char *t;

  if ((scratch = safe_strdup (str)) == NULL)
    return hints;

  for (t = strtok (scratch, " ,.:\"()<>\n"); t;
       t = strtok (NULL, " ,.:\"()<>\n"))
    {
      if (strlen (t) > 3)
        hints = mutt_add_list (hints, t);
    }
  
  FREE (scratch);
  return hints;
}




/* Convert LIST into a pattern string suitable to be passed to GPGME.
   We need to convert spaces in an item into a '+' and '%' into
   "%25". */
static char *list_to_pattern (LIST *list)
{
  LIST *l;
  char *pattern, *p;
  const char *s;
  size_t n;

  n = 0;
  for(l=list; l; l = l->next)
    {
      for(s = l->data; *s; s++)
        {
          if (*s == '%')
            n += 2;
          n++;
        }
      n++; /* delimiter or end of string */
    }
  n++; /* make sure to allocate at least one byte */
  pattern = p = safe_calloc (1,n);
  for(l=list; l; l = l->next)
    {
      s = l->data;
      if (*s)
        {
          if (l != list)
            *p++ = ' ';
          for(s = l->data; *s; s++)
            {
              if (*s == '%')
                {
                  *p++ = '%';
                  *p++ = '2';
                  *p++ = '5';
                }
              else if (*s == '+')
                {
                  *p++ = '%';
                  *p++ = '2';
                  *p++ = 'B';
                }
              else if (*s == ' ')
                *p++ = '+';
              else
                *p++ = *s;
            }
        }
    }
  *p = 0;
  return pattern;
}

/* Return a list of keys which are candidates for the selection.
   Select by looking at the HINTS list. */
static crypt_key_t *get_candidates (LIST * hints, unsigned int app, int secret)
{
  crypt_key_t *db, *k, **kend;
  char *pattern;
  GpgmeError err;
  GpgmeCtx ctx;
  GpgmeKey key;
  int idx;
  const char *uid;

  pattern = list_to_pattern (hints);
  if (!pattern)
    return NULL;
  
  err = gpgme_new (&ctx);
  if (err) 
    {
      mutt_error ("gpgme_new failed: %s", gpgme_strerror (err));
      FREE (pattern);
      return NULL;
    }

  db = NULL;
  kend = &db;
  
  if ((app & APPLICATION_PGP))
    {
      err = gpgme_op_keylist_start (ctx, pattern, secret);
      if (err) 
        {
          mutt_error ("gpgme_op_keylist_start failed: %s",
                      gpgme_strerror (err));
          gpgme_release (ctx);
          FREE (pattern);
          return NULL;
        }

      while (!(err = gpgme_op_keylist_next (ctx, &key)) )
        {
          unsigned int flags = 0;
          
          if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_CAN_ENCRYPT, NULL, 0))
            flags |= KEYFLAG_CANENCRYPT;
          if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_CAN_SIGN, NULL, 0))
            flags |= KEYFLAG_CANSIGN;

          if (!flags)
            { /* Bug in gpg.  Capabilities are not listed for secret
                 keys.  Try to deduce hem from the algorithm. */
              
              switch  (gpgme_key_get_ulong_attr (key,GPGME_ATTR_ALGO,NULL,0))
                {
                case 1:
                  flags |= KEYFLAG_CANENCRYPT;
                  flags |= KEYFLAG_CANSIGN;
                  break;
                case 16:
                  flags |= KEYFLAG_CANENCRYPT;
                  break;
                case 17:
                  flags |= KEYFLAG_CANSIGN;
                  break;
                }
            }
          
          for (idx=0; (uid = gpgme_key_get_string_attr (key, GPGME_ATTR_USERID,
                                                        NULL, idx)); idx++)
            {
              k = safe_calloc (1, sizeof *k);
              k->kobj = key;
              k->idx = idx;
              k->uid = uid;
              k->flags = flags;
              *kend = k;
              kend = &k->next;
            }
        }
      if (err != GPGME_EOF)
        mutt_error ("gpgme_op_keylist_next failed: %s", gpgme_strerror (err));
      gpgme_op_keylist_end (ctx);
    }

  if ((app & APPLICATION_SMIME))
    {
      /* and now look for x509 certificates */
      gpgme_set_protocol (ctx, GPGME_PROTOCOL_CMS);
      err = gpgme_op_keylist_start (ctx, pattern, 0);
      if (err) 
        {
          mutt_error ("gpgme_op_keylist_start failed: %s",
                      gpgme_strerror (err));
          gpgme_release (ctx);
          FREE (pattern);
          return NULL;
        }

      while (!(err = gpgme_op_keylist_next (ctx, &key)) )
        {
          unsigned int flags = KEYFLAG_ISX509;
          
          if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_CAN_ENCRYPT, NULL, 0))
            flags |= KEYFLAG_CANENCRYPT;
          if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_CAN_SIGN, NULL, 0))
            flags |= KEYFLAG_CANSIGN;
          
          for (idx=0; (uid = gpgme_key_get_string_attr (key, GPGME_ATTR_USERID,
                                                        NULL, idx)); idx++)
            {
              k = safe_calloc (1, sizeof *k);
              k->kobj = key;
              k->idx = idx;
              k->uid = uid;
              k->flags = flags;
              *kend = k;
              kend = &k->next;
            }
        }
      if (err != GPGME_EOF)
        mutt_error ("gpgme_op_keylist_next failed: %s", gpgme_strerror (err));
      gpgme_op_keylist_end (ctx);
    }

  gpgme_release (ctx);
  FREE (pattern);
  return db;
}


crypt_key_t *crypt_getkeybyaddr (ADDRESS * a, short abilities)
{
  ADDRESS *r, *p;
  LIST *hints = NULL;

  int weak    = 0;
  int invalid = 0;
  int multi   = 0;
  int this_key_has_strong;
  int this_key_has_weak;
  int this_key_has_invalid;
  int match;
  unsigned int app = APPLICATION_PGP|APPLICATION_SMIME;

  crypt_key_t *keys, *k;
  crypt_key_t *the_valid_key = NULL;
  crypt_key_t *matches = NULL;
  crypt_key_t **matches_endp = &matches;
  
  if (a && a->mailbox)
    hints = crypt_add_string_to_hints (hints, a->mailbox);
  if (a && a->personal)
    hints = crypt_add_string_to_hints (hints, a->personal);

  mutt_message (_("Looking for keys matching \"%s\"..."), a->mailbox);
  keys = get_candidates (hints, app, (abilities & KEYFLAG_CANSIGN) );

  mutt_free_list (&hints);
  
  if (!keys)
    return NULL;
  
  dprint (5, (debugfile, "crypt_getkeybyaddr: looking for %s <%s>.",
	      a->personal, a->mailbox));

  for (k = keys; k; k = k->next)
    {
      dprint (5, (debugfile, "  looking at key: %s `%.15s'\n",
                  crypt_keyid (k), k->uid));
      
      if (abilities && !(k->flags & abilities))
        {
          dprint (5, (debugfile, "  insufficient abilities: Has %x, want %x\n",
                      k->flags, abilities));
          continue;
        }

      this_key_has_weak    = 0;	/* weak but valid match   */
      this_key_has_invalid = 0;   /* invalid match          */
      this_key_has_strong  = 0;	/* strong and valid match */
      match                = 0;   /* any match 		  */

      r = rfc822_parse_adrlist (NULL, k->uid);
      for (p = r; p; p = p->next)
        {
          int validity = crypt_id_matches_addr (a, p, k);
              
          if (validity & CRYPT_KV_MATCH)	/* something matches */
            match = 1;

          /* is this key a strong candidate? */
          if ((validity & CRYPT_KV_VALID)
              && (validity & CRYPT_KV_STRONGID) 
              && (validity & CRYPT_KV_ADDR))
            {
              if (the_valid_key && the_valid_key != k)
                multi             = 1;
              the_valid_key       = k;
              this_key_has_strong = 1;
            }
          else if ((validity & CRYPT_KV_MATCH)
                   && !(validity & CRYPT_KV_VALID))
            this_key_has_invalid = 1;
          else if ((validity & CRYPT_KV_MATCH) 
                   && (!(validity & CRYPT_KV_STRONGID)
                       || !(validity & CRYPT_KV_ADDR)))
            this_key_has_weak    = 1;
        }
      rfc822_free_address (&r);
      
      if (match)
        {
          crypt_key_t *tmp;

          if (!this_key_has_strong && this_key_has_invalid)
            invalid = 1;
          if (!this_key_has_strong && this_key_has_weak)
            weak = 1;

          *matches_endp = tmp = crypt_copy_key (k);
          matches_endp = &tmp->next;
        }
    }
  
  crypt_free_key (&keys);
  
  if (matches)
    {
      if (the_valid_key && !multi && !weak 
          && !(invalid && option (OPTPGPSHOWUNUSABLE)))
        {	
          /* 
           * There was precisely one strong match on a valid ID, there
           * were no valid keys with weak matches, and we aren't
           * interested in seeing invalid keys.
           * 
           * Proceed without asking the user.
           */
          k = crypt_copy_key (the_valid_key);
        }
      else 
        {
          /* 
           * Else: Ask the user.
           */
          k = crypt_select_key (matches, a, NULL,
                                APPLICATION_PGP|APPLICATION_SMIME);
        }
      crypt_free_key (&matches);
    }
  else 
    k = NULL;
  
  return k;
}


crypt_key_t *crypt_getkeybystr (char *p, short abilities,
                                unsigned int app)
{
  LIST *hints = NULL;
  crypt_key_t *keys;
  crypt_key_t *matches = NULL;
  crypt_key_t **matches_endp = &matches;
  crypt_key_t *k;
  int match;

  mutt_message (_("Looking for keys matching \"%s\"..."), p);
  
  hints = crypt_add_string_to_hints (hints, p);
  keys = get_candidates (hints, app, (abilities & KEYFLAG_CANSIGN));
  mutt_free_list (&hints);

  if (!keys)
    return NULL;
  
  for (k = keys; k; k = k->next)
    {
      if (abilities && !(k->flags & abilities))
        continue;

      match = 0;
      dprint (5, (debugfile, "crypt_getkeybystr: matching \"%s\" against key %s, \"%s\": ",  p, crypt_keyid (k), k->uid));

      if (!*p
          || !mutt_strcasecmp (p, crypt_keyid (k))
          || (!mutt_strncasecmp (p, "0x", 2)
              && !mutt_strcasecmp (p + 2, crypt_keyid (k)))
          || (option (OPTPGPLONGIDS)
              && !mutt_strncasecmp (p, "0x", 2) 
              && !mutt_strcasecmp (p + 2, crypt_keyid (k) + 8))
          || mutt_stristr (k->uid, p))
        {
          crypt_key_t *tmp;

          dprint (5, (debugfile, "match.\n"));

          *matches_endp = tmp = crypt_copy_key (k);
          matches_endp = &tmp->next;
        }
    }
  
  crypt_free_key (&keys);
  
  if (matches)
    {
      k = crypt_select_key (matches, NULL, p, app);
      crypt_free_key (&matches);
      return k;
    }
  
  return NULL;
}


#endif /* HAVE_GPGME */
